package main

import (
    "os"
    "path"
    "time"
    "bytes"
    "errors"
    "strings"
    _ "embed"
    "os/exec"
    "encoding/json"
    "nexus-common/vm"
    "nexus-common/params"
    "nexus-common/buildversion"
    e "nexus-client/exec"
    u "nexus-client/utils"
    g "nexus-client/globals"
    "nexus-client/version"
    "nexus-client/defaults"
    "nexus-client/cmdLogin"
    "fyne.io/fyne/v2"
    "fyne.io/fyne/v2/app"
    "fyne.io/fyne/v2/theme"
    "fyne.io/fyne/v2/widget"
    "fyne.io/fyne/v2/container"
    "github.com/go-resty/resty/v2"
    "github.com/go-playground/validator/v10"
)

const (
    windowTitle = "nexus-exam"
)

var (
    //go:embed nexus_exam_user.val
    nexus_exam_user string

    //go:embed nexus_exam_pwd.val
    nexus_exam_pwd string

    token string
    certPath string
    exitFn func()
)

func exit(code int) {
    if exitFn != nil {
        exitFn()
    }
    os.Exit(code)
}

func attachVM(parent fyne.Window, hostname, cert, pwd string) {
    client := g.GetInstance().Client
    host := g.GetInstance().Host

    p := &params.VMAttachCreds{ Pwd: pwd }
    resp, err := client.R().SetBody(p).Post(host+"/vms/spicecreds")
    if err != nil {
        errorPopup(parent, "Failed attaching to VM (code 4)")
        return
    }

    if resp.IsSuccess() {
        var creds vm.VMSpiceCredentialsSerialized
        if err := json.Unmarshal(resp.Body(), &creds); err != nil {
            errorPopup(parent, "Failed attaching to VM (code 5)")
            return
        }
        if err := validator.New(validator.WithRequiredStructEnabled()).Struct(creds); err != nil {
            errorPopup(parent, "Failed attaching to VM (code 6)")
            return
        }
        go func(creds vm.VMSpiceCredentialsSerialized) {
            _, err := e.RunRemoteViewer(hostname, cert, creds.Name, creds.SpicePort, creds.SpicePwd, true)
            if err != nil {
                errorPopup(parent, "Failed attaching to VM (code 7)")
                return
            }
        } (creds)
    } else {
        type msg struct {
            Message string  `json:"message"`
        }
        var m msg
        if err := json.Unmarshal(resp.Body(), &m); err != nil {
            errorPopup(parent, "Failed attaching to VM (code 8)")
            return
        }
        errorPopup(parent, m.Message)
        // errorPopup(parent, "Error: "+resp.Status()+": "+resp.String())
        return
    }
}

func abortWindow(msg string) {
	a := app.New()
    a.Settings().SetTheme(theme.LightTheme())
	win := a.NewWindow(windowTitle)
    win.SetOnClosed(func() {
        win.Close()
		exit(1)
    })
	label := widget.NewLabel("FATAL: "+msg)
	button := widget.NewButton("Quit", func() {
        win.Close()
		exit(1)
	})
    content := container.NewPadded(container.NewVBox(label, button))
	win.SetContent(content)
    win.ShowAndRun()
}

func errorPopup(win fyne.Window, msg string) {
	var modal *widget.PopUp
	modal = widget.NewModalPopUp(
		container.NewVBox(
			widget.NewLabel("Error: "+msg),
			widget.NewButton("Close", func() { modal.Hide() })),
		win.Canvas())
	modal.Show()
}

func hypervisorCheck() {
    cmd := exec.Command("systemd-detect-virt")
    var out bytes.Buffer
    cmd.Stdout = &out
    cmd.Run()
    // Returns "none\n" if running on a physical system (no VM/container)
    if out.String() != "none\n" {
        abortWindow("nexus-exam only works on a physical system!")
    }
}

// Recurrently obtains a new JWT token so that the user session doesn't expire.
func refreshToken(parent fyne.Window) {
    client := g.GetInstance().Client
    host := g.GetInstance().Host

    for {
        resp, err := client.R().Get(host+"/token/refresh")
        if err != nil {
            errorPopup(parent, "Failed refreshing token (code 1)")
        } else {
            if resp.IsSuccess() {
                type Response struct {
                    Token string
                }
                var response Response
                err = json.Unmarshal(resp.Body(), &response)
                if err != nil {
                    errorPopup(parent, "Failed refreshing token (code 2)")
                }
                token = response.Token
            } else {
                // errorPopup(parent, resp.Status()+": "+resp.String())
                errorPopup(parent, "Failed refreshing token (code 3)")
            }
        }
        time.Sleep(4*time.Hour)
    }
}

func run() int {
    var appname = path.Base(os.Args[0])
    clientVersion := version.Get()
    u.PrintlnErr(appname+" version "+clientVersion.String()+" (commit "+buildversion.GitCommit()+")")
    u.PrintlnErr("Built on "+buildversion.BuildDate())

    hypervisorCheck()

    if err := e.CheckRemoteViewer(); err != nil {
        abortWindow(err.Error())
    }

    serverEnvVar, found := os.LookupEnv(g.ENV_NEXUS_SERVER)
    if !found {
        serverEnvVar = defaults.NexusServer
    }

    var err error  // necessary for certPath below to be ref as the same variable

    certPath, found = os.LookupEnv(g.ENV_NEXUS_CERT)
    if !found {
        certPath, err = defaults.CreateCert()
        if err != nil {
            abortWindow("Failed creating certificate from embedded certificate!")
        }
        // No embedded certificate, exit with a information message.
        if certPath == "" {
            abortWindow("environment variable \""+g.ENV_NEXUS_CERT+"\" is not set!")
        }
        exitFn = func() { os.Remove(certPath) }

        // This thread acts as a signal handler for SIGINT or SIGTERM.
        // When one of these signals is received, the temporary certificate file is deleted.
        // Without this "handler", the temporary certificate file wouldn't be deleted.
        go func() {
            u.WaitForSignals()
            exit(1)
        }()
    }

    parts := strings.Split(serverEnvVar, ":")
    hostname := parts[0]

    client := resty.New()
    client.SetRootCertificate(certPath)
    host := "https://"+serverEnvVar

    g.Init(hostname, host, certPath, client)

    client.SetTimeout(10*time.Second)

    // Checks the client version is compatible with the server's API.
    // if !cmdVersion.CheckServerCompatibility("nexus-exam") {
    //     abortWindow("client version is incompatible with server!")
    // }

    // Logins and obtains a JWT token.
    token, err = cmdLogin.GetToken(nexus_exam_user, nexus_exam_pwd)
    if err != nil {
        abortWindow("Failed obtaining token (network issue?)")
    }

    client.SetHeader("Content-Type", "application/json")
    client.SetHeader("Authorization", "Bearer "+token)

    app := app.New()
    app.Settings().SetTheme(theme.LightTheme())
    win := app.NewWindow(windowTitle)

    pwdEntry := widget.NewEntry()
    pwdEntry.Password = true

    pwdEntry.Validator = func(val string) error {
        if len(val) < 8 {
            return errors.New("Please enter at least 8 characters")
        }
        return nil
    }   

    label := widget.NewLabel("Enter password to connect to your VM:")

    form := &widget.Form{
        Items: []*widget.FormItem {
            {Text: "Password", Widget: pwdEntry},
        },
        OnSubmit: func() {
            attachVM(win, hostname, certPath, pwdEntry.Text)
        },
        SubmitText: "Connect",
    }

    win.SetContent(container.NewPadded(container.NewVBox(label, form)))
    win.Resize(fyne.NewSize(600,200))

    go refreshToken(win)

    win.ShowAndRun()
    return 0
}

func main() {
    exit(run())
}
